/*
 * label.c
 *    SE-PostgreSQL security label management
 *
 * Copyright (C) 2006-2010, NEC Corporation
 *                          KaiGai Kohei <kaigai@ak.jp.nec.com>
 */
#include "postgres.h"

#include "access/relscan.h"
#include "access/xact.h"
#include "catalog/pg_attribute.h"
#include "catalog/pg_class.h"
#include "catalog/pg_database.h"
#include "catalog/pg_largeobject_metadata.h"
#include "catalog/pg_namespace.h"
#include "catalog/pg_proc.h"
#include "catalog/pg_seclabel.h"
#include "catalog/pg_type.h"
#include "commands/dbcommands.h"
#include "miscadmin.h"
#include "sepgsql/sepgsql.h"
#include "utils/fmgroids.h"
#include "utils/lsyscache.h"
#include "utils/rel.h"
#include "utils/syscache.h"
#include "utils/tqual.h"
#include <fnmatch.h>	// for workaround hack
#include <selinux/selinux.h>

/*
 * guc parameter to turn on/off mcstrans
 */
bool sepgsql_mcstrans;

sepgsql_sid_t
sepgsql_move_secid(Oid dst_relid, sepgsql_sid_t ssid)
{
	char		   *label;
	sepgsql_sid_t	dsid = { .relid = dst_relid,
							 .secid = InvalidOid };

	label = seclabelRawOutput(ssid.relid, ssid.secid);
	if (label)
	{
		dsid.secid = seclabelRawInput(dsid.relid, label);

		pfree(label);
	}
	return dsid;
}

static sepgsql_sid_t
get_default_secid_with_database(Oid relOid, Oid databaseOid, uint16 tclass)
{
	HeapTuple		tuple;
	sepgsql_sid_t	datsid;

	tuple = SearchSysCache1(DATABASEOID,
							ObjectIdGetDatum(databaseOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for database: %u", databaseOid);

	datsid.relid = DatabaseRelationId;
	datsid.secid = HeapTupleGetSecid(tuple);

	ReleaseSysCache(tuple);

	return sepgsql_client_create_secid(datsid, tclass, relOid);
}

static sepgsql_sid_t
get_default_secid_with_schema(Oid relOid, Oid namespaceOid, uint16 tclass)
{
	HeapTuple		tuple;
	sepgsql_sid_t	nspsid;

	tuple = SearchSysCache1(NAMESPACEOID,
							ObjectIdGetDatum(namespaceOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for namespace: %u", namespaceOid);

	nspsid.relid = NamespaceRelationId;
	nspsid.secid = HeapTupleGetSecid(tuple);

	ReleaseSysCache(tuple);

	return sepgsql_client_create_secid(nspsid, tclass, relOid);
}

static sepgsql_sid_t
get_default_secid_with_table(Oid relOid, Oid tableOid, uint16 tclass)
{
	HeapTuple		tuple;
	sepgsql_sid_t	tblsid;

	tuple = SearchSysCache1(RELOID,
							ObjectIdGetDatum(tableOid));
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for relation: %u", tableOid);

	tblsid.relid = RelationRelationId;
	tblsid.secid = HeapTupleGetSecid(tuple);

	ReleaseSysCache(tuple);

	return sepgsql_client_create_secid(tblsid, tclass, relOid);
}

sepgsql_sid_t
sepgsql_get_default_database_secid(Oid templateOid)
{
	return get_default_secid_with_database(DatabaseRelationId,
										   templateOid,
										   SEPG_CLASS_DB_DATABASE);
}

sepgsql_sid_t
sepgsql_get_default_schema_secid(Oid databaseOid)
{
	return get_default_secid_with_database(NamespaceRelationId,
										   databaseOid,
										   SEPG_CLASS_DB_SCHEMA);
}

sepgsql_sid_t
sepgsql_get_default_table_secid(Oid namespaceOid)
{
	return get_default_secid_with_schema(RelationRelationId,
										 namespaceOid,
										 SEPG_CLASS_DB_TABLE);
}

sepgsql_sid_t
sepgsql_get_default_sequence_secid(Oid namespaceOid)
{
	return get_default_secid_with_schema(RelationRelationId,
										 namespaceOid,
										 SEPG_CLASS_DB_SEQUENCE);
}

sepgsql_sid_t
sepgsql_get_default_view_secid(Oid namespaceOid)
{
	return get_default_secid_with_schema(RelationRelationId,
										 namespaceOid,
										 SEPG_CLASS_DB_VIEW);
}

sepgsql_sid_t
sepgsql_get_default_proc_secid(Oid namespaceOid)
{
	return get_default_secid_with_schema(ProcedureRelationId,
										 namespaceOid,
										 SEPG_CLASS_DB_PROCEDURE);
}

sepgsql_sid_t
sepgsql_get_default_column_secid(Oid tableOid)
{
	return get_default_secid_with_table(AttributeRelationId,
										tableOid,
										SEPG_CLASS_DB_COLUMN);
}

sepgsql_sid_t
sepgsql_get_default_tuple_secid(Oid tableOid)
{
	return get_default_secid_with_table(tableOid,
										tableOid,
										SEPG_CLASS_DB_TUPLE);
}

sepgsql_sid_t
sepgsql_get_default_blob_secid(Oid databaseOid)
{
	return get_default_secid_with_database(LargeObjectMetadataRelationId,
										   databaseOid,
										   SEPG_CLASS_DB_BLOB);
}

Oid
sepgsql_get_default_secid(Relation rel, HeapTuple tuple)
{
	Oid				namespaceId;
	sepgsql_sid_t	nsid
		= { .relid = RelationGetRelid(rel), .secid = InvalidOid };

	switch (RelationGetRelid(rel))
	{
		case DatabaseRelationId:
		case RelationRelationId:
		case AttributeRelationId:
			elog(WARNING, "Bug? a new tuple without security id on \"%s\"",
				 RelationGetRelationName(rel));
			break;

		case NamespaceRelationId:
			nsid = sepgsql_get_default_schema_secid(MyDatabaseId);
			break;

		case ProcedureRelationId:
			namespaceId = ((Form_pg_proc) GETSTRUCT(tuple))->pronamespace;
			nsid = sepgsql_get_default_proc_secid(namespaceId);
			break;

		case LargeObjectMetadataRelationId:
			nsid = sepgsql_get_default_blob_secid(MyDatabaseId);
			break;

		default:
			nsid = sepgsql_get_default_tuple_secid(RelationGetRelid(rel));
			break;
	}
	return nsid.secid;
}

/*
 * a workaround implementation until libselinux/refpolicy don't
 * support db_schema or other object classes.
 */
static struct {
	uint16		tclass;
	char	   *pattern;
	char	   *context;
} initial_label_catalog[] = {
	{SEPG_CLASS_DB_DATABASE,		"*",
	 "system_u:object_r:sepgsql_db_t:s0"},
	{SEPG_CLASS_DB_SCHEMA,			"*.*",
	 "system_u:object_r:sepgsql_db_t:s0"},
	{SEPG_CLASS_DB_TABLE,			"*.pg_catalog.*",
	 "system_u:object_r:sepgsql_sysobj_t:s0"},
	{SEPG_CLASS_DB_TABLE,			"*.*.*",
	 "system_u:object_r:sepgsql_table_t:s0"},
	{SEPG_CLASS_DB_VIEW,			"*.*.*",
	 "system_u:object_r:sepgsql_db_t:s0"},
	{SEPG_CLASS_DB_SEQUENCE,		"*.*.*",
	 "system_u:object_r:sepgsql_db_t:s0"},
	{SEPG_CLASS_DB_PROCEDURE,		"*.pg_catalog.*",
	 "system_u:object_r:sepgsql_proc_exec_t:s0"},
	{SEPG_CLASS_DB_PROCEDURE,		"*.*.*",
	 "system_u:object_r:sepgsql_user_proc_exec_t:s0"},
	{SEPG_CLASS_DB_COLUMN,			"*.pg_catalog.*.*",
	 "system_u:object_r:sepgsql_sysobj_t:s0"},
	{SEPG_CLASS_DB_COLUMN,			"*.*.*.*",
	 "system_u:object_r:sepgsql_table_t:s0"},
	{SEPG_CLASS_DB_TUPLE,			"*.pg_catalog.*",
	 "system_u:object_r:sepgsql_sysobj_t:s0"},
	{SEPG_CLASS_DB_TUPLE,			"*.*.*",
	 "system_u:object_r:sepgsql_table_t:s0"},
	{SEPG_CLASS_DB_BLOB,			"*.*",
	 "system_u:object_r:sepgsql_blob_t:s0"},
	{0, NULL, NULL},
};

static char *
lookup_init_catalog(uint16 tclass, const char *name)
{
	int	i;

	for (i = 0; initial_label_catalog[i].pattern; i++)
	{
		if (initial_label_catalog[i].tclass == tclass &&
			fnmatch(initial_label_catalog[i].pattern, name, 0) == 0)
			return initial_label_catalog[i].context;
	}
	elog(ERROR, "no valid initial security context for %s (tclass=%d)",
		 name, tclass);
	return NULL;	/* for compiler quiet */
}

static char *
lookup_init_tuple_label(Oid relOid, HeapTuple tuple)
{
	Oid			relNsp = get_rel_namespace(relOid);
	char		namebuf[NAMEDATALEN * 3 + 10];

	snprintf(namebuf, sizeof(namebuf), "%s.%s.%s",
			 get_database_name(MyDatabaseId),
			 get_namespace_name(relNsp),
			 get_rel_name(relOid));

	return lookup_init_catalog(SEPG_CLASS_DB_TUPLE, namebuf);
}

static char *
lookup_init_database_label(HeapTuple tuple)
{
	Form_pg_database	datForm = (Form_pg_database) GETSTRUCT(tuple);

	return lookup_init_catalog(SEPG_CLASS_DB_DATABASE,
							   NameStr(datForm->datname));
}

static char *
lookup_init_schema_label(HeapTuple tuple)
{
	Form_pg_namespace	nspForm = (Form_pg_namespace) GETSTRUCT(tuple);
	char		namebuf[NAMEDATALEN * 2 + 10];

	snprintf(namebuf, sizeof(namebuf), "%s.%s",
			 get_database_name(MyDatabaseId),
			 NameStr(nspForm->nspname));

	return lookup_init_catalog(SEPG_CLASS_DB_SCHEMA, namebuf);
}

static char *
lookup_init_relation_label(HeapTuple tuple)
{
	Form_pg_class		classForm = (Form_pg_class) GETSTRUCT(tuple);
	const char		   *relName = NameStr(classForm->relname);
	Oid					relNsp = classForm->relnamespace;
	char			   *seclabel;
	char				namebuf[NAMEDATALEN * 3 + 10];

	switch (classForm->relkind)
	{
		case RELKIND_RELATION:
			snprintf(namebuf, sizeof(namebuf), "%s.%s.%s",
					 get_database_name(MyDatabaseId),
					 get_namespace_name(relNsp), relName);
			seclabel = lookup_init_catalog(SEPG_CLASS_DB_TABLE, namebuf);
			break;

		case RELKIND_SEQUENCE:
			snprintf(namebuf, sizeof(namebuf), "%s.%s.%s",
					 get_database_name(MyDatabaseId),
					 get_namespace_name(relNsp), relName);
			seclabel = lookup_init_catalog(SEPG_CLASS_DB_SEQUENCE, namebuf);
			break;

		case RELKIND_VIEW:
			snprintf(namebuf, sizeof(namebuf), "%s.%s.%s",
					 get_database_name(MyDatabaseId),
					 get_namespace_name(relNsp), relName);
			seclabel = lookup_init_catalog(SEPG_CLASS_DB_VIEW, namebuf);
			break;

		case RELKIND_INDEX: {
			HeapTuple	tbltup;
			HeapTuple	indtup;
			Oid			tblOid;
			Oid			indOid = HeapTupleGetOid(tuple);

			indtup = SearchSysCache1(INDEXRELID, ObjectIdGetDatum(indOid));
			if (!HeapTupleIsValid(indtup))
				elog(ERROR, "cache lookup failed for index %u", indOid);

			tblOid = ((Form_pg_index) GETSTRUCT(indtup))->indrelid;
			tbltup = SearchSysCache1(RELOID, ObjectIdGetDatum(tblOid));
			if (!HeapTupleIsValid(tbltup))
				elog(ERROR, "cache lookup failed for relation %u", tblOid);

			seclabel = lookup_init_relation_label(tbltup);

			ReleaseSysCache(tbltup);
			ReleaseSysCache(indtup);

			break;
		}
		case RELKIND_TOASTVALUE: {
			HeapTuple	tbltup;
			Oid			tblOid;

			/*
			 * XXX - we assume all the toast relation's name is
			 * "pg_toast_%u", and the "%u" shall be replaced by OID
			 * of the relation which owns the toast relation
			 */
			tblOid = strtoul(relName + 9, NULL, 10);

			tbltup = SearchSysCache1(RELOID,
									 ObjectIdGetDatum(tblOid));
			if (!HeapTupleIsValid(tbltup))
				elog(ERROR, "cache lookup failed for relation %u", tblOid);

			seclabel = lookup_init_relation_label(tbltup);

			ReleaseSysCache(tbltup);

			break;
		}
		case RELKIND_COMPOSITE_TYPE: {
			Oid			typOid = classForm->reltype;
			HeapTuple	typtup;

			typtup = SearchSysCache1(TYPEOID,
									 ObjectIdGetDatum(typOid));
			if (!HeapTupleIsValid(typtup))
				elog(ERROR, "cache lookup failed for type %u", typOid);

			seclabel = lookup_init_tuple_label(TypeRelationId, typtup);

			ReleaseSysCache(typtup);

			break;
		}
		default:
			elog(ERROR, "unexpected relkind %c of \"%s\"",
				 classForm->relkind, relName);
			seclabel = NULL;	/* compiler quiet */
			break;
	}
	return seclabel;
}

static char *
lookup_init_attribute_label(HeapTuple tuple)
{
	Form_pg_attribute	attForm = (Form_pg_attribute) GETSTRUCT(tuple);
	Oid			tblOid = attForm->attrelid;
	char	   *seclabel;
	char		namebuf[NAMEDATALEN * 4 + 10];

	if (get_rel_relkind(tblOid) == RELKIND_RELATION)
	{
		Oid		tblNsp = get_rel_namespace(tblOid);

		snprintf(namebuf, sizeof(namebuf), "%s.%s.%s.%s",
				 get_database_name(MyDatabaseId),
				 get_namespace_name(tblNsp),
				 get_rel_name(tblOid),
				 NameStr(attForm->attname));

		seclabel = lookup_init_catalog(SEPG_CLASS_DB_COLUMN, namebuf);
	}
	else
	{
		HeapTuple	tbltup;

		tbltup = SearchSysCache1(RELOID, ObjectIdGetDatum(tblOid));
		if (!HeapTupleIsValid(tbltup))
			elog(ERROR, "cache lookup failed for relation %u", tblOid);

		seclabel = lookup_init_relation_label(tbltup);

		ReleaseSysCache(tbltup);
	}
	return seclabel;
}

static char *
lookup_init_procedure_label(HeapTuple tuple)
{
	Form_pg_proc	proForm = (Form_pg_proc) GETSTRUCT(tuple);
	Oid				proNsp = proForm->pronamespace;
	char			namebuf[NAMEDATALEN * 3 + 10];

	snprintf(namebuf, sizeof(namebuf), "%s.%s.%s",
			 get_database_name(MyDatabaseId),
			 get_namespace_name(proNsp),
			 NameStr(proForm->proname));

	return lookup_init_catalog(SEPG_CLASS_DB_PROCEDURE, namebuf);
}

static char *
lookup_init_largeobject_label(HeapTuple tuple)
{
	char		namebuf[NAMEDATALEN + 20];

	snprintf(namebuf, sizeof(namebuf), "%s.%u",
			 get_database_name(MyDatabaseId),
			 HeapTupleGetOid(tuple));

	return lookup_init_catalog(SEPG_CLASS_DB_BLOB, namebuf);
}

void
sepgsql_initial_labeling(void)
{
	Relation		classRel;
	SysScanDesc		classScan;
	ScanKeyData		classSkey;
	HeapTuple		classTup;
	Relation		rel;
	HeapScanDesc	scan;
	HeapTuple		oldtup;
	HeapTuple		newtup;

	Assert(IsBootstrapProcessingMode());

	StartTransactionCommand();

	classRel = heap_open(RelationRelationId, AccessShareLock);

	ScanKeyInit(&classSkey,
				Anum_pg_class_relhassecids,
				BTEqualStrategyNumber, F_BOOLEQ,
				BoolGetDatum(true));

	classScan = systable_beginscan(classRel, InvalidOid, false,
								   SnapshotNow, 1, &classSkey);

	while (HeapTupleIsValid(classTup = systable_getnext(classScan)))
	{
		Oid		relOid = HeapTupleGetOid(classTup);

		Assert(((Form_pg_class) GETSTRUCT(classTup))->relhassecids);

		rel = heap_open(relOid, RowExclusiveLock);

		scan = heap_beginscan(rel, SnapshotNow, 0, NULL);

		while (HeapTupleIsValid(oldtup = heap_getnext(scan, ForwardScanDirection)))
		{
			char	   *label;
			Oid			secid;

			switch (relOid)
			{
				case DatabaseRelationId:
					label = lookup_init_database_label(oldtup);
					break;

				case NamespaceRelationId:
					label = lookup_init_schema_label(oldtup);
					break;

				case RelationRelationId:
					label = lookup_init_relation_label(oldtup);
					break;

				case AttributeRelationId:
					label = lookup_init_attribute_label(oldtup);
					break;

				case ProcedureRelationId:
					label = lookup_init_procedure_label(oldtup);
					break;

				case LargeObjectMetadataRelationId:
					label = lookup_init_largeobject_label(oldtup);
					break;

				default:
					label = lookup_init_tuple_label(relOid, oldtup);
					break;
			}
			/*
			 * inplace-updating
			 */
			newtup = heap_copytuple(oldtup);

			secid = seclabelTransInput(relOid, label);

			HeapTupleSetSecid(newtup, secid);

			heap_inplace_update(rel, newtup);

			heap_freetuple(newtup);
		}
		heap_endscan(scan);
		
		heap_close(rel, RowExclusiveLock);
	}
	systable_endscan(classScan);

	heap_close(classRel, AccessShareLock);

	CommitTransactionCommand();
}

char *
sepgsql_mcstrans_in(char *trans_label)
{
	security_context_t	raw_label;
	security_context_t	result;

	if (!sepgsql_mcstrans)
		return trans_label;

	if (selinux_trans_to_raw_context(trans_label, &raw_label) < 0)
		ereport(ERROR,
				(errcode(ERRCODE_INTERNAL_ERROR),
				 errmsg("SELinux: unable to translate \"%s\"", trans_label)));
	PG_TRY();
	{
		result = pstrdup(raw_label);
	}
	PG_CATCH();
	{
		freecon(raw_label);
		PG_RE_THROW();
	}
	PG_END_TRY();
	freecon(raw_label);

	return result;
}

char *
sepgsql_mcstrans_out(char *raw_label)
{
	security_context_t	trans_label;
	security_context_t	result;

	if (!sepgsql_mcstrans)
		return raw_label;

	if (selinux_raw_to_trans_context(raw_label, &trans_label) < 0)
		ereport(ERROR,
				(errcode(ERRCODE_INTERNAL_ERROR),
				 errmsg("SELinux: unable to translate \"%s\"", raw_label)));
	PG_TRY();
	{
		result = pstrdup(trans_label);
	}
	PG_CATCH();
	{
		freecon(trans_label);
		PG_RE_THROW();
	}
	PG_END_TRY();
	freecon(trans_label);

	return result;
}

char *
sepgsql_rawlabel_in(char *label)
{
	if (!label || security_check_context_raw(label) < 0)
		ereport(ERROR,
				(errcode(ERRCODE_INVALID_NAME),
				 errmsg("invalid security context \"%s\"", label)));
	return label;
}

char *
sepgsql_rawlabel_out(char *label)
{
	if (!label || security_check_context_raw(label) < 0)
	{
		security_context_t	unlabeled_label;

		if (security_get_initial_context_raw("unlabeled",
											 &unlabeled_label) < 0)
			ereport(ERROR,
					(errcode(ERRCODE_INTERNAL_ERROR),
					 errmsg("unable to get unlabeled security context")));
		PG_TRY();
		{
			label = pstrdup(unlabeled_label);
		}
		PG_CATCH();
		{
			freecon(unlabeled_label);
			PG_RE_THROW();
		}
		PG_END_TRY();
		freecon(unlabeled_label);
	}
	return label;
}
